/* -*- js-indent-level: 8 -*- */
/*
 * L.Control.ColumnGroup
*/

/* global app */

/*
	This file is Calc only. This adds a section for grouped columns in Calc.
	When user selects some columns and groups them using "Data->Group and Outline->Group" menu path, this section is added into
	sections list of CanvasSectionContainer. See _addRemoveGroupSections in file CalcTileLayer.js

	This class is an extended version of "CanvasSectionObject".
*/

L.Control.ColumnGroup = L.Control.GroupBase.extend({
	name: L.CSections.ColumnGroup.name,
	anchor: ['top', [L.CSections.CornerGroup.name, 'right', 'left']],
	position: [0, 0], // This section's myTopLeft is placed according to corner group section if exists, if not, this is placed at (0, 0).
	size: [0, 0], // No initial width is necessary. Width will be expanded. Height is computed inside update function.
	expand: ['left', 'right'], // Expand horizontally.
	processingOrder: L.CSections.ColumnGroup.processingOrder,
	drawingOrder: L.CSections.ColumnGroup.drawingOrder,
	zIndex: L.CSections.ColumnGroup.zIndex,
	interactable: true,
	sectionProperties: {},

	// This function is called by CanvasSectionContainer when the section is added to the sections list.
	onInitialize: function () {
		this._map = L.Map.THIS;
		this._groups = null;

		// group control styles
		this._groupHeadSize = Math.round(12 * app.dpiScale);
		this._levelSpacing = app.roundedDpiScale;

		this._map.on('sheetgeometrychanged', this.update, this);
		this._map.on('viewrowcolumnheaders', this.update, this);
		this._createFont();
		this.update();
		this.isRemoved = false;
	},

	update: function () {
		if (this.isRemoved) // Prevent calling while deleting the section. It causes errors.
			return;

		this._sheetGeometry = this._map._docLayer.sheetGeometry;
		this._groups = Array(this._sheetGeometry.getColumnGroupLevels());

		// Calculate width on the fly.
		this.size[1] = this._computeSectionHeight();

		this._cornerHeaderWidth = this.containerObject.getSectionWithName(L.CSections.CornerHeader.name).size[0];

		this._splitPos = this._map._docLayer._splitPanesContext.getSplitPos();

		this._collectGroupsData(this._sheetGeometry.getColumnGroupsDataInView());
	},

	// This returns the required height for the section.
	_computeSectionHeight: function () {
		return this._levelSpacing + (this._groupHeadSize + this._levelSpacing) * (this._groups.length + 1);
	},

	getRelativeX: function (docPos) {
		if (docPos <= this._splitPos.x) {
			return docPos - this.documentTopLeft[0] + this._cornerHeaderWidth;
		}
		else {
			// max here is to prevent encroachment of the fixed pane-area.
			return Math.max(docPos - this.documentTopLeft[0], this._splitPos.x) + this._cornerHeaderWidth;
		}
	},

	drawGroupControl: function (group) {
		var startX = this.getRelativeX(group.startPos);
		var startY = this._levelSpacing + (this._groupHeadSize + this._levelSpacing) * group.level;
		var endX = group.endPos + this._cornerHeaderWidth - this.documentTopLeft[0];

		if (startX > this._cornerHeaderWidth) {
			// draw head
			this.context.fillStyle = this.backgroundColor;
			this.context.fillRect(startX, startY, this._groupHeadSize, this._groupHeadSize);
			this.context.strokeStyle = 'black';
			this.context.lineWidth = app.dpiScale;
			this.context.strokeRect(startX + 0.5, startY + 0.5, this._groupHeadSize, this._groupHeadSize);
		}

		if (!group.hidden && endX > startX) {
			//draw tail
			startY += this._groupHeadSize * 0.5;
			this.context.strokeStyle = 'black';
			this.context.lineWidth = app.dpiScale;
			this.context.beginPath();
			this.context.moveTo(startX + this._groupHeadSize + 0.5, startY + 0.5);
			this.context.lineTo(endX - app.roundedDpiScale + 0.5, startY + 0.5);
			this.context.lineTo(endX - app.roundedDpiScale + 0.5, startY + this._groupHeadSize / 2);
			this.context.stroke();
			startY -= this._groupHeadSize * 0.5;
			if (startX > this._cornerHeaderWidth) {
				// draw '-'
				this.context.moveTo(startX + this._groupHeadSize * 0.25, startY + this._groupHeadSize * 0.5 + 0.5);
				this.context.lineTo(startX + this._groupHeadSize * 0.75 + app.roundedDpiScale, startY + this._groupHeadSize * 0.5 + 0.5);
				this.context.stroke();
			}
		}
		else if (startX > this._cornerHeaderWidth) {
			// draw '+'
			this.context.beginPath();
			this.context.moveTo(startX + this._groupHeadSize * 0.25, startY + this._groupHeadSize * 0.5 + 0.5);
			this.context.lineTo(startX + this._groupHeadSize * 0.75 + app.roundedDpiScale, startY + this._groupHeadSize * 0.5 + 0.5);

			this.context.moveTo(startX + this._groupHeadSize * 0.50 + 0.5, startY + this._groupHeadSize * 0.25);
			this.context.lineTo(startX + this._groupHeadSize * 0.50 + 0.5, startY + this._groupHeadSize * 0.75 + app.roundedDpiScale);

			this.context.stroke();
		}
	},

	drawLevelHeader: function (level) {
		var ctx = this.context;
		var ctrlHeadSize = this._groupHeadSize;
		var levelSpacing = this._levelSpacing;

		var startX = Math.round((this._cornerHeaderWidth - ctrlHeadSize) * 0.5);
		var startY = levelSpacing + (ctrlHeadSize + levelSpacing) * level;

		ctx.strokeStyle = 'black';
		ctx.lineWidth = app.dpiScale;
		ctx.strokeRect(startX + 0.5, startY + 0.5, ctrlHeadSize, ctrlHeadSize);
		// draw level number
		ctx.fillStyle = this._textColor;
		ctx.font = this._getFont();
		ctx.textAlign = 'center';
		ctx.textBaseline = 'middle';
		ctx.fillText(level + 1, startX + (ctrlHeadSize / 2), startY + (ctrlHeadSize / 2) + 2 * app.dpiScale);
	},

	// Handle user interaction.
	_updateOutlineState: function (group) {
		var state = group.hidden ? 'visible' : 'hidden'; // we have to send the new state
		var payload = 'outlinestate type=column' + ' level=' + group.level + ' index=' + group.index + ' state=' + state;
		app.socket.sendMessage(payload);
	},

	// When user clicks somewhere on the section, onMouseClick event is called by CanvasSectionContainer.
	// Clicked point is also given to handler function. This function finds the clicked header.
	findClickedLevel: function (point) {
		if (point[0] < this._cornerHeaderWidth) {
			var index = (point[1] / this.size[1]) * 100; // Percentage.
			var levelPercentage = (1 / (this._groups.length + 1)) * 100; // There is one more button than the number of levels.
			index = Math.floor(index / levelPercentage);
			return index;
		}
		else {
			return -1;
		}
	},

	findClickedGroup: function (point) {
		for (var i = 0; i < this._groups.length; i++) {
			if (this._groups[i]) {
				for (var group in this._groups[i]) {
					if (Object.prototype.hasOwnProperty.call(this._groups[i], group)) {
						var group_ = this._groups[i][group];
						var startX = this.getRelativeX(group_.startPos);
						var startY = this._levelSpacing + (this._groupHeadSize + this._levelSpacing) * group_.level;
						var endX = startX + this._groupHeadSize;
						var endY = group_.endPos + this._cornerHeaderWidth - this.documentTopLeft[1];
						if (point[0] > startX && point[0] < endX && point[1] > startY && point[1] < endY) {
							return group_;
						}
					}
				}
			}
		}
		return null;
	},

	// Users can double click on group tails.
	findTailsGroup: function (point) {
		for (var i = 0; i < this._groups.length; i++) {
			if (this._groups[i]) {
				for (var group in this._groups[i]) {
					if (Object.prototype.hasOwnProperty.call(this._groups[i], group)) {
						var group_ = this._groups[i][group];
						var startX = this.getRelativeX(group_.startPos);
						var startY = this._levelSpacing + (this._groupHeadSize + this._levelSpacing) * group_.level;
						var endX = group_.endPos + this._cornerHeaderWidth - this.documentTopLeft[0];
						var endY = startY + this._groupHeadSize;

						if (point[0] > startX && point[0] < endX && point[1] > startY && point[1] < endY) {
							return group_;
						}
					}
				}
			}
		}
	},

	onRemove: function () {
		this.isRemoved = true;
		this.containerObject.getSectionWithName(L.CSections.ColumnHeader.name).position[1] = 0;
		this.containerObject.getSectionWithName(L.CSections.CornerHeader.name).position[1] = 0;
	},
});

L.control.columnGroup = function (options) {
	return new L.Control.ColumnGroup(options);
};